from naas_drivers.driver import OutDriver
import requests


class Ifttt(OutDriver):
    def connect(self, key):
        self._key = key
        self.connected = True
        return self

    def send(self, event, data=None):
        self.check_connect()
        url = f"https://maker.ifttt.com/trigger/{event}/with/key/{self._key}"
        r = requests.post(
            url=url,
            json=data,
        )
        r.raise_for_status()
        return r



from naas_drivers.driver import OutDriver
import hashlib
import os
import requests


class Taggun(OutDriver):
    """
    Naas Driver for Taggun.io

    Taggun offers affordable Receipt/invoice OCR,
        Input: a pdf or image of a receipt
        Output: a dict containing parsed data from the image, and a corresponding
            confidence level for the extraction
    """

    _key = None
    _fileName = None
    _filePath = None
    _url = None
    _incognito = None
    _hash = None

    data = None

    def connect(self, key, mode="simple", incognito=False):
        """
        Initialize all the necessary elements to make an api request

        Args:
            key (string): Taggun API key
            mode (string): Either simple or verbose. default simple
            incognite (bool): If true, taggun will not store a copy of the
                uploaded file for reinforcement learning. default false
        """
        self._key = key
        self.connected = True
        self._url = f"https://api.taggun.io/api/receipt/v1/{mode}/file"
        self._incognito = incognito
        return self

    def hash_file(self, file):
        """
        Create a hash of the file to be sent. Taggun allows user-submitted data
        for reinforcement learning. This requires a unique identifier for the file.
        Instead of generating a UUID, which would need to be stored, use a hash
        of the file itself as a unique identifier

        Returns:
            hash of the file
        """
        with open(file, "rb") as f:
            self._hash = hashlib.sha256(f.read()).hexdigest()
        return self._hash

    def send(self, file):
        """
        Dispatch the api request to Taggun and receive parsed data

        Args:
        file (string): Path to the file being sent for parsing

        Returns:
            parsed data as a dict from Taggun
        """
        self._fileName = os.path.basename(file)
        self._filePath = file
        self.check_connect()
        self.hash_file(self._filePath)

        with open(self._filePath, "rb") as f:
            headers = {"apikey": self._key}
            files = {
                "file": (
                    self._fileName,
                    f,
                    "application/pdf",
                ),  # content-type for the file
                # other optional parameters for Taggun API (eg: incognito, refresh, ipAddress, language)
                "incognito": (
                    None,  # set filename to none for optional parameters
                    self._incognito,
                ),  # value for the parameters
                "referenceID": (
                    None,  # set filename to none for optional parameters
                    self._hash,  # value for the parameters
                ),
            }
            response = requests.post(self._url, files=files, headers=headers)
            self.data = response.json()
        return self.data



import sys
import logging
from typing import Dict, List, Optional
from pandas import DataFrame

import snowflake.connector
from snowflake.connector.errors import ProgrammingError
from naas_drivers.driver import InDriver, OutDriver


global snowflake_instance


class Snowflake(InDriver, OutDriver):
    """
    Driver for Snowflake Data Warehouse. Enables to connect to the SF account and execute queries.
    There is no need to log in and use Worksheets. Connection instance is enough to run any command.

    Basic drill recommendation:
        - `snowflake.connect()` - to connect to the account with given credentials
        - `snowflake.set_environment()` - to set up SF working environment for the whole session
        - `snowflake.execute()` or `snowflake.query_pd()` - to run any command you want
            (and eventually get the results in a nice Pandas DataFrame)

    For more information, please refer to functions' documentation.

    Moreover, there are 2 extra session's objects accessible: `cursor` and `connection` that are raw
        `snowflake-python-connector` objects. If you need, you can refer to Snowflake docs and perform extra tasks that
        aren't (yet) implemented in this driver.

    If the functionality you're looking for is not implemented yet, please create an issue in
        (naas-drivers repository)[https://github.com/jupyter-naas/drivers/tree/main],
        we would be more than happy to help.
    """

    def __init__(self):

        self._connection = None
        self._cursor = None

        # Creating Snowflake objects
        self._database = Database(self)
        self._file_format = FileFormat(self)
        self._role = Role(self)
        self._schema = Schema(self)
        self._stage = Stage(self)
        self._storage_integration = StorageIntegration(self)
        self._warehouse = Warehouse(self)

        # Providing global instance to use after loading snowflake driver module
        global snowflake_instance
        snowflake_instance = self

    @staticmethod
    def instance():
        return snowflake_instance

    @property
    def connection(self):
        return self._connection

    @property
    def cursor(self):
        return self._cursor

    @property
    def database(self):
        return self._database

    @property
    def file_format(self):
        return self._file_format

    @property
    def role(self):
        return self._role

    @property
    def schema(self):
        return self._schema

    @property
    def stage(self):
        return self._stage

    @property
    def storage_integration(self):
        return self._storage_integration

    @property
    def warehouse(self):
        return self._warehouse

    def connect(
        self,
        account: str,
        username: str,
        password: str,
        warehouse: str = "",
        database: str = "",
        schema: str = "",
        role: str = "",
    ) -> None:
        """
        Connects to Snowflake account with given credentials.
        Connection is established and both `connection` and `cursor` are being set up
        @param account: SF account identifier, can be fetched from login URL,
            e.g. <account_identifier>.snowflakecomputing.com
        @param username: SF account username
        @param password: SF account password
        @param warehouse: SF warehouse to set up while creating a connection
        @param database: SF database to set up while creating a connection
        @param schema: SF schema to set up while creating a connection
        @param role: SF role to set up while creating a connection
        """
        self._connection = snowflake.connector.connect(
            account=account, user=username, password=password
        )
        self._cursor = self._connection.cursor()
        self.set_environment(warehouse, database, schema, role)
        self.connected = True

    def execute(self, sql: str, n: int = 10, silent: bool = False) -> Optional[Dict]:
        """
        Execute passed command. Could be anything, starting from DQL query, and ending with DDL commands
        @param sql: command/query to execute
        @param n: query result length limit
        @param silent: whether to return result dictionary with multiple information or not (run in silent mode)
        @return: dictionary containing query information outcome (results, columns_metadata, and sql statement)
        """
        res = self._cursor.execute(sql)

        if silent:
            return None
        else:
            return {
                "results": res.fetchall() if n == -1 else res.fetchmany(n),
                "description": res.description,
                "statement": sql,
            }

    def query_pd(self, sql: str, n: int = 10) -> DataFrame:
        """
        Query data and return results in the form of pandas.DataFrame
        @param sql: query to execute
        @param n: query result length limit
        @return: pandas.DataFrame table containing query outcome
        """
        res = self.execute(sql, n)

        # TODO: Apply dtypes mapping from ResultMetadata objects
        return DataFrame(
            res["results"],
            columns=[result_metadata.name for result_metadata in res["description"]],
        )

    def copy_into(
        self,
        table_name: str,
        source_stage: str,
        transformed_columns: str = "",
        files: List[str] = None,
        regex_pattern: str = "",
        file_format_name: str = "",
        validation_mode: str = "",
        silent: bool = False,
        **kwargs,
    ) -> Optional[Dict]:
        """
        Copy data from stage to Snowflake table
        As this function contains a lot of nuances, please refer to the documentation:
            https://docs.snowflake.com/en/sql-reference/sql/copy-into-table.html
        @param table_name: the name of the table into which data is loaded. Optionally, can specify namespace too
            (i.e., <database_name>.<schema_name> or <schema_name> if database is already selected)
        @param source_stage: internal or external location where the files containing data to be loaded are staged
            caution: it can be also a SELECT statement so data from stage is transformed
        @param transformed_columns: if `source_stage` is a SELECT statement,
            specifies an explicit set of fields/columns (separated by commas) to load from the staged data files
        @param files: a list of one or more files names (in an array) to be loaded
        @param regex_pattern: a regular expression pattern for filtering files from the output
        @param file_format_name: the format of the data files to load (so far only name is accepted as an input)
        @param validation_mode: instructs the COPY command to validate the data files
            instead of loading them into the specified table.
            Caution: does not support COPY statements that transform data during a load.
            If applied along with transformation SELECT statement, it will throw error on Snowflake side
        @param silent: whether to run in silent mode (see `Snowflake.execute()`)
        @return: result dictionary (see: `Snowflake.execute()`)
        """
        statement = "COPY INTO" f" {table_name}"

        # If applicable, then data load with transformation has been scheduled
        # If not, it's a standard data load
        if source_stage.upper().strip().startswith("SELECT"):
            statement += (
                f"{' ' + transformed_columns if transformed_columns != '' else ''}"
                f" FROM ({source_stage})"
            )
        else:
            statement += f" FROM {source_stage}"

        files_string = (
            ",".join([f"'{file}'" for file in files]) if files is not None else ""
        )
        statement += f"{f' ({files_string})' if files_string != '' else ''}"

        statement += (
            f"{f' PATTERN = {regex_pattern}' if regex_pattern != '' else ''}"
            f"{f' FILE FORMAT = (FORMAT_NAME = {file_format_name})' if file_format_name != '' else ''}"
            f"{f' VALIDATION_MODE = {validation_mode}' if validation_mode != '' else ''}"
        )

        # looping through kwargs for extra arguments passed in statement
        # while executing final command, Snowflake will do the validation
        for key, value in kwargs.items():
            statement += f" {key} = {value}"

        return self.execute(statement, n=1, silent=silent)

    def close_connection(self) -> None:
        """
        Closes a connection to Snowflake account, resets all the internal parameters
        """
        self._cursor.close()
        self._connection.close()

        self._cursor = None
        self._connection = None

    def get_environment(self) -> Dict:
        """
        Returns current session environment that consists of:
            - role
            - database
            - schema
            - warehouse
        @return: dictionary of current session environment elements
        """

        return {
            "role": self._role.get_current(),
            "database": self._database.get_current(),
            "schema": self._schema.get_current(),
            "warehouse": self._warehouse.get_current(),
        }

    def set_environment(
        self, warehouse: str = "", database: str = "", schema: str = "", role: str = ""
    ) -> None:
        """
        Tries to set Snowflake environment in bulk: warehouse, database, schema, and role
        @param warehouse: warehouse to set in the environment
        @param database: database to set in the environment
        @param schema: schema to set in the environment
        @param role: role to set in the environment
        """
        warehouse, database, schema, role = (
            warehouse.strip(),
            database.strip(),
            schema.strip(),
            role.strip(),
        )
        try:
            if warehouse != "":
                self._warehouse.use(warehouse)
            if database != "":
                self._database.use(database)
            if schema != "":
                self._schema.use(schema)
            if role != "":
                self._role.use(role)
        except ProgrammingError as pe:
            logging.error(f"Error while setting SF environment. More on that: {pe}")
            sys.exit()


class Database:
    def __init__(self, snowflake_driver: Snowflake):
        self.__snowflake_driver = snowflake_driver

    def use(self, database_name: str, silent: bool = False) -> Optional[Dict]:
        """
        Sets particular database for a session
        @param database_name: name of the database to use
        @param silent: whether to run in silent mode (see `Snowflake.execute()`)
        @return: result dictionary (see: `Snowflake.execute()`)
        """
        statement = "USE" f" DATABASE {database_name}"

        return self.__snowflake_driver.execute(statement, n=1, silent=silent)

    def get_current(
        self,
    ) -> str:
        """
        Returns the name of the database in use by the current session
        @return: result dictionary (see: `Snowflake.execute()`)
        """
        statement = "SELECT CURRENT_DATABASE()"

        return self.__snowflake_driver.execute(statement, n=1)["results"][0][0]

    def create(
        self, database_name: str, or_replace: bool = False, silent: bool = False
    ) -> Optional[Dict]:
        """
        Executes command to create a Snowflake database with a given name
        @param database_name: database name to create
        @param or_replace: replace schema if exists
        @param silent: whether to run in silent mode (see `Snowflake.execute()`)
        @return: result dictionary (see: `Snowflake.execute()`)
        """
        statement = (
            "CREATE"
            f"{' OR REPLACE' if or_replace else ''}"
            f" DATABASE {database_name}"
        )

        return self.__snowflake_driver.execute(statement, n=1, silent=silent)

    def drop(
        self, database_name: str, if_exists: bool = False, silent: bool = False
    ) -> Optional[Dict]:
        """
        Executes command to drop a Snowflake database with a given name
        @param database_name: database name to drop
        @param if_exists: adds `IF EXISTS` statement to a command
        @param silent: whether to run in silent mode (see `Snowflake.execute()`)
        @return: result dictionary (see: `Snowflake.execute()`)
        """
        statement = (
            "DROP DATABASE" f"{' IF EXISTS' if if_exists else ''}" f" {database_name}"
        )

        return self.__snowflake_driver.execute(statement, n=1, silent=silent)


class Schema:
    def __init__(self, snowflake_driver: Snowflake):
        self.__snowflake_driver = snowflake_driver

    def use(self, schema_name: str, silent: bool = False) -> Optional[Dict]:
        """
        Sets particular schema for a session
        @param schema_name: name of the schema to use
        @param silent: whether to run in silent mode (see `Snowflake.execute()`)
        @return: result dictionary (see: `Snowflake.execute()`)
        """
        statement = "USE" f" SCHEMA {schema_name}"

        return self.__snowflake_driver.execute(statement, n=1, silent=silent)

    def get_current(
        self,
    ) -> str:
        """
        Returns the name of the schema in use by the current session
        @return: result dictionary (see: `Snowflake.execute()`)
        """
        statement = "SELECT CURRENT_SCHEMA()"

        return self.__snowflake_driver.execute(statement, n=1)["results"][0][0]

    def create(
        self, schema_name: str, or_replace: bool = False, silent: bool = False
    ) -> Optional[Dict]:
        """
        Executes command to create a Snowflake schema with a given name
        @param schema_name: schema name to create
        @param or_replace: replace schema if exists
        @param silent: whether to run in silent mode (see `Snowflake.execute()`)
        @return: result dictionary (see: `Snowflake.execute()`)
        """
        statement = (
            "CREATE" f"{' OR REPLACE' if or_replace else ''}" f" SCHEMA {schema_name}"
        )

        return self.__snowflake_driver.execute(statement, n=1, silent=silent)

    def drop(
        self, schema_name: str, if_exists: bool = False, silent: bool = False
    ) -> Optional[Dict]:
        """
        Executes command to drop a Snowflake schema with a given name
        @param schema_name: schema name to drop
        @param if_exists: adds `IF EXISTS` statement to a command
        @param silent: whether to run in silent mode (see `Snowflake.execute()`)
        @return: result dictionary (see: `Snowflake.execute()`)
        """
        statement = (
            "DROP SCHEMA" f"{' IF EXISTS' if if_exists else ''}" f" {schema_name}"
        )

        return self.__snowflake_driver.execute(statement, n=1, silent=silent)


class FileFormat:
    def __init__(self, snowflake_driver: Snowflake):
        self.__snowflake_driver = snowflake_driver
        self.__AVAILABLE_FORMAT_TYPES = ["CSV", "JSON", "AVRO", "ORC", "PARQUET", "XML"]

    def create(
        self,
        file_format_name: str,
        file_format_type: str,
        or_replace: bool = False,
        if_not_exists: bool = False,
        silent: bool = False,
        **kwargs,
    ) -> Optional[Dict]:
        """
        Executes command to create a Snowflake file format with a given name
        @param file_format_name: file format name to create
        @param file_format_type: type of the file format to be created
        @param or_replace: replace file format if exists
        @param if_not_exists: create object if it doesn't exist so far
        @param silent: whether to run in silent mode (see `Snowflake.execute()`)
        @param kwargs: additional arguments to be passed to the statement
            so far validation is on the Snowflake engine side
        @return: result dictionary (see: `Snowflake.execute()`)
        """
        file_format_type = file_format_type.upper().strip()
        if file_format_type not in self.__AVAILABLE_FORMAT_TYPES:
            raise ValueError(
                f"File Format type `{file_format_type}` not available for now"
            )

        statement = (
            "CREATE"
            f"{' OR REPLACE' if or_replace else ''}"
            f" FILE FORMAT{' IF NOT EXISTS' if if_not_exists else ''} {file_format_name}"
            f" TYPE = {file_format_type}"
        )

        # looping through kwargs for extra arguments passed in statement
        # while executing final command, Snowflake will do the validation
        for key, value in kwargs.items():
            statement += f" {key} = {value}"

        return self.__snowflake_driver.execute(statement, n=1, silent=silent)

    def drop(
        self, file_format_name: str, if_exists: bool = False, silent: bool = False
    ) -> Optional[Dict]:
        """
        Executes command to drop a Snowflake file format with a given name
        @param file_format_name: file format name to drop
        @param if_exists: adds `IF EXISTS` statement to a command
        @param silent: whether to run in silent mode (see `Snowflake.execute()`)
        @return: result dictionary (see: `Snowflake.execute()`)
        """
        statement = (
            "DROP FILE FORMAT"
            f"{' IF EXISTS' if if_exists else ''}"
            f" {file_format_name}"
        )

        return self.__snowflake_driver.execute(statement, n=1, silent=silent)


class Stage:
    def __init__(self, snowflake_driver: Snowflake):
        self.__snowflake_driver = snowflake_driver

    def create(
        self,
        stage_name: str,
        or_replace: bool = False,
        is_temporary: bool = False,
        if_not_exists: bool = False,
        file_format_name: str = "",
        silent: bool = False,
        **kwargs,
    ) -> Optional[Dict]:
        """
        Executes command to create a Snowflake stage with a given name
        @param stage_name: stage name to create
        @param or_replace: replace file format if exists
        @param is_temporary: create a temporary stage
        @param if_not_exists: create object if it doesn't exist so far
        @param file_format_name: file format name to use while creating a stage
        @param silent: whether to run in silent mode (see `Snowflake.execute()`)
        @param kwargs: additional arguments to be passed to the statement
            so far validation is on the Snowflake engine side
        @return: result dictionary (see: `Snowflake.execute()`)
        """
        statement = (
            "CREATE"
            f"{' OR REPLACE' if or_replace else ''}"
            f"{' TEMPORARY' if is_temporary else ''}"
            f" STAGE"
            f"{' IF NOT EXISTS' if if_not_exists else ''}"
            f" {stage_name}"
            f"{f' FILE_FORMAT = {file_format_name}' if file_format_name != '' else ''}"
        )

        # looping through kwargs for extra arguments passed in statement
        # while executing final command, Snowflake will do the validation
        for key, value in kwargs.items():
            statement += f" {key} = {value}"

        return self.__snowflake_driver.execute(statement, n=1, silent=silent)

    def drop(
        self, stage_name: str, if_exists: bool = False, silent: bool = False
    ) -> Optional[Dict]:
        """
        Executes command to drop a Snowflake stage with a given name
        @param stage_name: stage name to drop
        @param if_exists: adds `IF EXISTS` statement to a command
        @param silent: whether to run in silent mode (see `Snowflake.execute()`)
        @return: result dictionary (see: `Snowflake.execute()`)
        """
        statement = "DROP STAGE" f"{' IF EXISTS' if if_exists else ''}" f" {stage_name}"

        return self.__snowflake_driver.execute(statement, n=1, silent=silent)

    def put(
        self,
        filepath: str,
        internal_stage_name: str,
        parallel: int = 4,
        auto_compress: bool = True,
        source_compression: str = "AUTO_DETECT",
        overwrite: bool = False,
        silent: bool = False,
    ) -> Optional[Dict]:
        """
        Executes command to put data to Snowflake internal stage from a local machine
        @param filepath: local path to a file
        @param internal_stage_name: stage name where to put file a file
            Caution: when having problems with this param, please check whether you specify the proper
            internal stage name, including special signs: @, %, and ~
        @param parallel: number of threads to use for uploading files
        @param auto_compress: whether Snowflake uses gzip to compress files during upload
        @param source_compression: method of compression used on already-compressed files that are being staged
        @param overwrite: whether Snowflake overwrites an existing file with the same name during upload
        @param silent: whether to run in silent mode (see `Snowflake.execute()`)
        @return: result dictionary (see: `Snowflake.execute()`)
        """
        statement = (
            "PUT"
            f" '{filepath}'"
            f" {internal_stage_name}"
            f" PARALLEL = {parallel}"
            f" AUTO_COMPRESS = {'TRUE' if auto_compress else 'FALSE'}"
            f" SOURCE_COMPRESSION = {source_compression}"
            f" OVERWRITE = {'TRUE' if overwrite else 'FALSE'}"
        )

        return self.__snowflake_driver.execute(statement, n=1, silent=silent)

    def list(
        self, stage_name: str, regex_pattern: str = "", silent: bool = False
    ) -> Optional[Dict]:
        """
        Lists files that are inside in a particular stage
        @param stage_name: the location where the data files are staged
        @param regex_pattern: a regular expression pattern for filtering files from the output
        @param silent: whether to run in silent mode (see `Snowflake.execute()`)
        @return: result dictionary (see: `Snowflake.execute()`)
        """
        statement = (
            "LIST"
            f" {stage_name}"
            f"{f' PATTERN = {regex_pattern}' if regex_pattern != '' else ''}"
        )

        return self.__snowflake_driver.execute(statement, n=1, silent=silent)


class Role:
    def __init__(self, snowflake_driver: Snowflake):
        self.__snowflake_driver = snowflake_driver

    def use(self, role_name: str, silent: bool = False) -> Optional[Dict]:
        """
        Sets particular role for a session
        @param role_name: name of the role to use
        @param silent: whether to run in silent mode (see `Snowflake.execute()`)
        @return: result dictionary (see: `Snowflake.execute()`)
        """
        statement = "USE" f" ROLE {role_name}"

        return self.__snowflake_driver.execute(statement, n=1, silent=silent)

    def get_current(
        self,
    ) -> str:
        """
        Returns the name of the role in use by the current session
        @return: result dictionary (see: `Snowflake.execute()`)
        """
        statement = "SELECT CURRENT_ROLE()"

        return self.__snowflake_driver.execute(statement, n=1)["results"][0][0]


class StorageIntegration:
    def __init__(self, snowflake_driver: Snowflake):
        self.__snowflake_driver = snowflake_driver

    def create(
        self,
        storage_integration_name: str,
        storage_provider: str,
        storage_allowed_locations: List[str],
        or_replace: bool = False,
        if_not_exists: bool = False,
        enabled: bool = True,
        silent: bool = False,
        **kwargs,
    ) -> Optional[Dict]:
        """
        Executes command to create a Snowflake stage with a given name
        @param storage_integration_name: storage integration name to create
        @param storage_provider: cloud provider to create stage and fetch data from
        @param storage_allowed_locations: explicitly limits external stages
            that use the integration to reference one or more storage locations
        @param or_replace: replace file format if exists
        @param if_not_exists: create object if it doesn't exist so far
        @param enabled: specifies whether this storage integration is available for usage in stages
        @param silent: whether to run in silent mode (see `Snowflake.execute()`)
        @param silent: whether to run in silent mode (see `Snowflake.execute()`)
        @return: result dictionary (see: `Snowflake.execute()`)
        """
        storage_allowed_loccations_string = ",".join(
            [f"'{file}'" for file in storage_allowed_locations]
        )

        statement = (
            "CREATE"
            f"{' OR REPLACE' if or_replace else ''}"
            f" STORAGE INTEGRATION{' IF NOT EXISTS' if if_not_exists else ''} {storage_integration_name}"
            f" TYPE = EXTERNAL_STAGE"
            f" ENABLED = {'TRUE' if enabled else 'FALSE'}"
            f" STORAGE_PROVIDER = {storage_provider}"
            f" STORAGE_ALLOWED_LOCATIONS = ({storage_allowed_loccations_string})"
        )

        # looping through kwargs for extra arguments passed in statement
        # while executing final command, especially cloud-provider-specific parameters,
        # Snowflake will do the validation
        for key, value in kwargs.items():
            statement += f" {key} = {value}"

        return self.__snowflake_driver.execute(statement, n=1, silent=silent)

    def drop(
        self,
        storage_integration_name: str,
        if_exists: bool = False,
        silent: bool = False,
    ) -> Optional[Dict]:
        """
        Executes command to drop a Snowflake storage integration with a given name
        @param storage_integration_name: storage integration name to drop
        @param if_exists: adds `IF EXISTS` statement to a command
        @param silent: whether to run in silent mode (see `Snowflake.execute()`)
        @return: result dictionary (see: `Snowflake.execute()`)
        """
        statement = (
            "DROP STORAGE INTEGRATION"
            f"{' IF EXISTS' if if_exists else ''}"
            f" {storage_integration_name}"
        )

        return self.__snowflake_driver.execute(statement, n=1, silent=silent)


class Warehouse:
    def __init__(self, snowflake_driver: Snowflake):
        self.__snowflake_driver = snowflake_driver

    def use(self, warehouse_name: str, silent: bool = False) -> Optional[Dict]:
        """
        Sets particular warehouse for a session
        @param warehouse_name: name of the warehouse to use
        @param silent: whether to run in silent mode (see `Snowflake.execute()`)
        @return: result dictionary (see: `Snowflake.execute()`)
        """
        statement = "USE" f" WAREHOUSE {warehouse_name}"

        return self.__snowflake_driver.execute(statement, n=1, silent=silent)

    def get_current(
        self,
    ) -> str:
        """
        Returns the name of the warehouse in use by the current session
        @return: result dictionary (see: `Snowflake.execute()`)
        """
        statement = "SELECT CURRENT_WAREHOUSE()"

        return self.__snowflake_driver.execute(statement, n=1)["results"][0][0]



import requests
import pandas as pd


class Bazimo:
    def connect(self, email: str, password: str):

        # Connect to Bazimo
        url = "https://bazimo-api.azurewebsites.net/api/tenant/app/authentications/jwt"

        json = {
            "email": f"{email}",
            "password": f"{password}",
            "authenticationType": "bearer",
        }
        res = requests.post(url, json=json)
        res.raise_for_status()
        res_json = res.json()
        token = res_json.get("token")

        # Init token
        self.token = token

        # Init end point
        self.exports = Exports(self.token)

        # Set connexion to active
        self.connected = True
        return self


class Exports(Bazimo):
    def __init__(self, token):
        Bazimo.__init__(self)
        self.token = token

    def get(self, name):
        if name == "Locataires":
            scope = "1"
            header = 6
            sheet_name = "Locataires"
        if name == "Lots":
            scope = "2"
            header = 6
            sheet_name = "Lots"
        if name == "Baux":
            scope = "3"
            header = 6
            sheet_name = "Baux"
        if name == "Actifs":
            scope = "9"
            header = 6
            sheet_name = "Actifs"
        if name == "Factures":
            scope = "11"
            header = 5
            sheet_name = "Détail factures"

        # Request url
        url = (
            f"https://bazimo-api.azurewebsites.net/api/tenant/app/exports?scope={scope}"
        )
        headers = {"Authorization": f"bearer {self.token}"}
        res = requests.get(url, headers=headers)
        res.raise_for_status()

        # Read Excel
        df = pd.read_excel(res.content, sheet_name=sheet_name, header=header)
        return df



import pandas as pd
import requests
import time
import urllib
from datetime import datetime
import secrets
import pydash as _pd
import naas
from naas_drivers.tools.emailbuilder import EmailBuilder
from naas_drivers.tools.naas_auth import NaasAuth
import json

emailbuilder = EmailBuilder()
naasauth = NaasAuth()

LINKEDIN_API = "https://3hz1hdpnlf.execute-api.eu-west-1.amazonaws.com/prod"
RELEASE_MESSAGE = (
    "Feature not release yet."
    "Please create or comment issue on Jupyter Naas Github: "
    "https://github.com/orgs/jupyter-naas/projects/4"
)
DATE_FORMAT = "%Y-%m-%d"
DATETIME_FORMAT = "%Y-%m-%d %H:%M:%S"
TIME_SLEEP = secrets.randbelow(5) + 5
HEADERS = {"Content-Type": "application/json"}
EMAIL_COOKIES = "⚠️ Naas.ai - Update your Linkedin cookies"


class LinkedIn:
    deprecated = True

    @staticmethod
    def get_user_email():
        email = None
        user = naasauth.connect().user.me()
        email = user.get("username")
        return email

    @staticmethod
    def email_linkedin_limit(email):
        content = {
            "header_naas": (
                "<a href='https://www.naas.ai/'>"
                "<img align='center' width='30%' target='_blank' style='border-radius:5px;'"
                "src='https://landen.imgix.net/jtci2pxwjczr/assets/5ice39g4.png?w=160'"
                "alt='Please provide more information.'/>"
                "</a>"
            ),
            "txt_0": emailbuilder.text(
                "Hi there,<br><br>"
                "Your LinkedIn cookies needs to be renewed.<br><br>"
                "Please go to naas and update them in your notebook 'Setup LinkedIn'.<br>"
            ),
            "button": emailbuilder.button(
                f"https://app.naas.ai/user/{email}/", "Go to Naas"
            ),
            "signature": "Naas Team",
            "footer": emailbuilder.footer_company(naas=True),
        }
        email_content = emailbuilder.generate(display="iframe", **content)
        return email_content

    @staticmethod
    def send_email_renewed_cookies():
        email = LinkedIn.get_user_email()
        email_content = LinkedIn.email_linkedin_limit(email)
        naas.notification.send(
            email_to=email, subject=EMAIL_COOKIES, html=email_content
        )

    @staticmethod
    def manage_api_error(res):
        if res.status_code != 200:
            if int(res.status_code) == 302:
                LinkedIn.send_email_renewed_cookies()
                raise requests.TooManyRedirects(res.status_code, res.text)
            else:
                raise BaseException(res.status_code, res.text)

    @staticmethod
    def get_activity_id(url):
        if "-activity-" in url:
            return url.split("-activity-")[-1].split("-")[0]
        if ":activity:" in url:
            return url.split(":activity:")[-1]

    def print_deprecated(self, new_funct):
        if self.deprected:
            print(f"This function is deprecated, please use {new_funct}")

    def get_profile_id(self, url):
        return url.rsplit("/in/")[-1].rsplit("/")[0]

    def get_profile_urn(self, url):
        lk_id = self.get_profile_id(url)
        res = requests.get(
            f"https://www.linkedin.com/voyager/api/identity/profiles/{lk_id}",
            cookies=self.cookies,
            headers=self.headers,
        )
        # Check if requests is successful
        try:
            res.raise_for_status()
            res_json = res.json()
            return (
                res_json.get("data", {})
                .get("entityUrn")
                .replace("urn:li:fs_profile:", "")
            )
        except requests.HTTPError as e:
            return e

    def get_birthdate(self, bd):
        if bd is None:
            return "No birthdate"
        bd_day = bd.get("day", "Day Unknown")
        bd_month = bd.get("month", "Month Unknown")
        bd_year = bd.get("year", "Year Unknown")
        return f"{bd_day}/{bd_month}/{bd_year}"

    def clear_occupation(self, occupation):
        if occupation is not None:
            occupation = occupation.strip().replace("\n", " ")
        return occupation

    def connect(
        self,
        li_at: str = None,
        jessionid: str = None,
    ):
        # Init lk attribute
        self.li_at = li_at
        self.jessionid = jessionid

        # Init cookies
        self.cookies = {"li_at": self.li_at, "JSESSIONID": f'"{self.jessionid}"'}

        # Init headers
        self.headers = {
            "X-Li-Lang": "en_US",
            "Accept": "application/vnd.linkedin.normalized+json+2.1",
            "Cache-Control": "no-cache",
            "csrf-Token": self.jessionid.replace('"', ""),
            "X-Requested-With": "XMLHttpRequest",
            "X-Restli-Protocol-Version": "2.0.0",
        }

        # Init end point
        self.profile = Profile(self.cookies, self.headers)
        self.network = Network(self.cookies, self.headers)
        self.invitation = Invitation(self.cookies, self.headers)
        self.message = Message(self.cookies, self.headers)
        self.post = Post(self.cookies, self.headers)
        self.event = Event(self.cookies, self.headers)
        self.company = Company(self.cookies, self.headers)

        # Set connexion to active
        self.connected = True
        return self


class Profile(LinkedIn):
    def __init__(self, cookies, headers):
        LinkedIn.__init__(self)
        self.cookies = cookies
        self.headers = headers

    def get_identity(self, profile_url=None, sleep=True):
        """
        Return an dataframe object with 15 columns:
        - FIRSTNAME
        - LASTNAME
        - SUMMARY
        - OCCUPATION
        - INDUSTRY_NAME
        - ADDRESS
        - REGION
        - COUNTRY
        - LOCATION
        - BIRTHDATE
        - PROFILE_ID
        - PROFILE_URL
        - PUBLIC_ID
        - BACKGROUND_PICTURE
        - PROFILE_PICTURE

        Parameters
        ----------
        profile_url: str:
            Profile URL from LinkedIn.
            Example : "https://www.linkedin.com/in/florent-ravenel/"
        """
        res_json = {}
        if profile_url is None:
            print("❌ No profile URL. Please enter a profile URL from LinkedIn")
            return res_json
        result = {}
        lk_public_id = self.get_profile_id(profile_url)
        req_url = (
            f"https://www.linkedin.com/voyager/api/identity/profiles/{lk_public_id}"
        )
        res = requests.get(req_url, cookies=self.cookies, headers=self.headers)
        # Raise error
        res.raise_for_status()
        # Parse json
        res_json = res.json()
        data = res_json.get("data", {})
        included = res_json.get("included", {})

        # Init var
        bg_pic_url = None
        profile_pic_url = None

        # Get data from included json
        if len(included) > 0:
            included = included[0]

            # Get background picture
            if included.get("backgroundImage"):
                background_url_end = None
                background_root = None
                background_artifacts = _pd.get(included, "backgroundImage.artifacts")
                if len(background_artifacts) > 0:
                    background_url_end = background_artifacts[
                        len(background_artifacts) - 1
                    ].get("fileIdentifyingUrlPathSegment")
                background_root = _pd.get(included, "backgroundImage.rootUrl")
                if background_url_end and background_root:
                    bg_pic_url = f"{background_root}{background_url_end}"
            # Get profile picture
            if included.get("picture"):
                profile_url_end = None
                profile_root = None
                profile_artifacts = _pd.get(included, "picture.artifacts")
                if len(profile_artifacts) > 0:
                    profile_url_end = profile_artifacts[len(profile_artifacts) - 1].get(
                        "fileIdentifyingUrlPathSegment"
                    )
                profile_root = _pd.get(included, "picture.rootUrl")
                if profile_root and profile_url_end:
                    profile_pic_url = f"{profile_root}{profile_url_end}"
        lk_id = data.get("entityUrn", "").replace("urn:li:fs_profile:", "")
        result = {
            "PROFILE_ID": lk_id,
            "PROFILE_URL": f"https://www.linkedin.com/in/{lk_id}",
            "PUBLIC_ID": lk_public_id,
            "FIRSTNAME": data.get("firstName"),
            "LASTNAME": data.get("lastName"),
            "SUMMARY": data.get("summary"),
            "OCCUPATION": data.get("headline"),
            "INDUSTRY_NAME": data.get("industryName"),
            "ADDRESS": data.get("address"),
            "REGION": data.get("geoLocationName"),
            "COUNTRY": data.get("geoCountryName"),
            "LOCATION": data.get("locationName"),
            "BIRTHDATE": self.get_birthdate(data.get("birthDateOn")),
            "BACKGROUND_PICTURE": bg_pic_url,
            "PROFILE_PICTURE": profile_pic_url,
        }
        if sleep:
            time.sleep(TIME_SLEEP)
        return pd.DataFrame([result])

    def get_network(self, profile_url=None, sleep=True):
        """
        Return an dataframe object with 7 columns:
        - PROFILE_ID
        - PROFILE_URL
        - PUBLIC_ID
        - DISTANCE
        - FOLLOWING
        - FOLLOWABLE
        - FOLLOWERS_COUNT

        Parameters
        ----------
        profile_url: str:
            Profile URL from LinkedIn.
            Example : "https://www.linkedin.com/in/florent-ravenel/"
        """
        res_json = {}
        if profile_url is None:
            print("❌ No profile URL. Please enter a profile URL from LinkedIn")
            return res_json
        result = {}
        lk_id = self.get_profile_id(profile_url)
        req_url = f"https://www.linkedin.com/voyager/api/identity/profiles/{lk_id}/networkinfo"
        res = requests.get(req_url, cookies=self.cookies, headers=self.headers)
        # Raise error
        res.raise_for_status()
        # Parse json
        res_json = res.json()
        data = res_json.get("data", {})
        profile_id = data.get("entityUrn", "").replace(
            "urn:li:fs_profileNetworkInfo:", ""
        )
        result = {
            "PROFILE_ID": profile_id,
            "PROFILE_URL": f"https://www.linkedin.com/in/{profile_id}",
            "PUBLIC_ID": lk_id,
            "DISTANCE": data.get("distance", {}).get("value"),
            "FOLLOWING": data.get("following"),
            "FOLLOWABLE": data.get("followable"),
            "FOLLOWERS_COUNT": data.get("followersCount"),
        }
        if sleep:
            time.sleep(TIME_SLEEP)
        return pd.DataFrame([result])

    def get_contact(self, profile_url=None, sleep=True):
        """
        Return an dataframe object with 11 columns:
        - PROFILE_ID
        - PROFILE_URL
        - PUBLIC_ID
        - EMAIL
        - CONNECTED_AT
        - BIRTHDATE
        - ADDRESS
        - TWITTER
        - PHONENUMBER
        - WEBSITES
        - INTERESTS

        Parameters
        ----------
        profile_url: str:
            Profile URL from LinkedIn.
            Example : "https://www.linkedin.com/in/florent-ravenel/"
        """
        res_json = {}
        if profile_url is None:
            print("❌ No profile URL. Please enter a profile URL from LinkedIn")
            return res_json
        result = {}
        lk_id = self.get_profile_id(profile_url)
        req_url = f"https://www.linkedin.com/voyager/api/identity/profiles/{lk_id}/profileContactInfo"
        res = requests.get(req_url, cookies=self.cookies, headers=self.headers)
        res.raise_for_status()
        # Parse json
        res_json = res.json()
        data = res_json.get("data", {})

        # Specific
        connected_at = data.get("connectedAt")
        if connected_at is not None:
            connected_at = datetime.fromtimestamp(int(str(connected_at)[:-3])).strftime(
                "%Y-%m-%d %H:%M:%S"
            )
        lk_phone = None
        lk_phones = data.get("phoneNumbers")
        if lk_phones is not None:
            for rows in lk_phones:
                if rows["type"] == "MOBILE":
                    lk_phone = rows["number"]
                    break
        lk_twiter = None
        lk_twiters = data.get("twitterHandles")
        if lk_twiters is not None:
            for rows in lk_twiters:
                lk_twiter = rows["name"]
                break
        lk_urls = ""
        lk_websites = data.get("websites")
        if lk_websites is not None:
            for rows in lk_websites:
                lk_url = rows["url"]
                lk_urls = f"{lk_urls}{lk_url}, "
        profile_id = data.get("entityUrn", "").replace("urn:li:fs_contactinfo:", "")
        result = {
            "PROFILE_ID": profile_id,
            "PROFILE_URL": f"https://www.linkedin.com/in/{profile_id}",
            "PUBLIC_ID": lk_id,
            "EMAIL": data.get("emailAddress"),
            "CONNECTED_AT": connected_at,
            "BIRTHDATE": self.get_birthdate(data.get("birthDateOn")),
            "ADDRESS": data.get("address"),
            "TWITTER": lk_twiter,
            "PHONENUMBER": lk_phone,
            "WEBSITES": lk_urls,
            #             "INTERESTS": data.get("interests"),
        }
        if sleep:
            time.sleep(TIME_SLEEP)
        return pd.DataFrame([result])

    def get_resume(self, profile_url=None, profile_urn=None):
        """
        Return an dataframe object with 13 columns:
        - PROFILE_ID                    object
        - PROFILE_URL                   object
        - FULL_NAME                     object
        - CATEGORY                      object
        - TITLE                         object
        - DATE_START                    object
        - DATE_END                      object
        - PLACE_ID                      object
        - PLACE                         object
        - FIELD                         object
        - LOCATION                      object
        - DESCRIPTION                   object
        - DATE_EXTRACT                  object

        Parameters
        ----------
        profile_url: str:
            Profile URL from LinkedIn.
            Example : "https://www.linkedin.com/in/florent-ravenel/"
        """
        res_json = {}
        if profile_url is None:
            print("❌ No profile URL. Please enter a profile URL from LinkedIn")
            return res_json
        if profile_urn is None:
            profile_urn = LinkedIn.get_profile_urn(self, profile_url)
            if profile_urn is None:
                return "Please enter a valid profile_url or profile_urn"
        req_url = f"{LINKEDIN_API}/profile/getResume?profile_urn={profile_urn}"
        res = requests.post(req_url, json=self.cookies, headers=HEADERS)
        res.raise_for_status()

        # Manage LinkedIn API errors
        LinkedIn.manage_api_error(res)

        # Get json result
        res_json = res.json()
        df = pd.DataFrame(res_json)
        return df.reset_index(drop=True)

    def get_posts_feed(
        self,
        profile_url,
        profile_id=None,
        count=1,
        limit=10,
        until={},
        sleep=True,
        pagination_token=None,
    ):
        """
        Return an dataframe object with 32 columns:
        - ACTIVITY_ID                   object
        - PAGINATION_TOKEN              object
        - PUBLISHED_DATE                object
        - AUTHOR_NAME                   object
        - AUTHOR_URL                    object
        - SUBDESCRIPTION                object
        - TITLE                         object
        - TEXT                          object
        - CHARACTER_COUNT               int64
        - TAGS                          object
        - TAGS_COUNT                    int64
        - EMOJIS                        object
        - EMOJIS_COUNT                  int64
        - LINKS                         object
        - LINKS_COUNT                   int64
        - PROFILE_MENTION               object
        - COMPANY_MENTION               object
        - CONTENT                       object
        - CONTENT_TITLE                 object
        - CONTENT_URL                   object
        - CONTENT_ID                    object
        - IMAGE_URL                     object
        - POLL_ID                       object
        - POLL_QUESTION                 object
        - POLL_RESULTS                  object
        - POST_URL                      object
        - VIEWS                         int64
        - COMMENTS                      int64
        - LIKES                         int64
        - SHARES                        int64
        - ENGAGEMENT_SCORE              float64
        - DATE_EXTRACT                  object

        Parameters
        ----------
        profile_url: str:
            Profile url from Linkedin.
            Example : "https://www.linkedin.com/in/florent-ravenel/"

        profile_id: str (default None):
            Linkedin unique profile id identifier
            Example : "ACoAABCNSioBW3YZHc2lBHVG0E_TXYWitQkmwog"

        count: int (default 1, max 100):
            Number of requests sent to LinkedIn API.
            (!) If count > 1, published date will not be returned.

        limit: int (default 10, unlimited=-1):
            Number of posts return by function. It will start with the most recent post.

        until: dict (default {})
            Dict to be set by end user to limit function:
            - key must a columns of the dataframe
            - value must exists in key columns
            Example : "{"POST_URL": "https://www.linkedin.com/posts/naas-ai_opensource-data-activity-6890025972754710529-akfv"

        sleep: boolean (default True):
            Sleeping time between function will be randomly between 3 to 5 seconds.

        pagination_token: str (default None):
            Token related to post used to start function from this post.
            If None, function starts from the last post.

        """
        # Get profile
        if profile_id is None:
            profile_id = LinkedIn.get_profile_urn(self, profile_url)
            if profile_id is None:
                return "Please enter a valid profile_url or profile_urn"
        # Until init
        until_check = False
        keys = []
        if isinstance(until, dict) and len(until) > 0:
            keys = [k for k, v in until.items()]
        # Loop init
        start = 0
        df = pd.DataFrame()
        if limit != -1 and count > limit:
            limit = count
        while True:
            if pagination_token is not None:
                req_url = f"{LINKEDIN_API}/profile/getPostsFeed?profile_id={profile_id}&count={count}&pagination_token={pagination_token}"
            else:
                req_url = f"{LINKEDIN_API}/profile/getPostsFeed?profile_id={profile_id}&count={count}"
            res = requests.post(req_url, json=self.cookies, headers=HEADERS)
            res.raise_for_status()

            # Manage LinkedIn API errors
            LinkedIn.manage_api_error(res)

            # Get json result
            res_json = res.json()
            if len(res_json) == 0:
                break
            # Get response in dataframe
            tmp_df = pd.DataFrame(res_json)

            # Check until limit
            for k in keys:
                v = until.get(k)
                if k in tmp_df.columns:
                    values = tmp_df[k].astype(str).unique().tolist()
                    if str(v) in values:
                        until_check = True
                        break
            # Get pagination token + update start
            pagination_token = tmp_df.loc[0, "PAGINATION_TOKEN"]

            # Concat dataframe
            df = pd.concat([df, tmp_df], axis=0)

            # Break if until condition is True
            if until_check:
                break
            # Time sleep to avoid linkedin ban
            if sleep:
                time.sleep(TIME_SLEEP)
            start += count
            if limit != -1 and start >= limit:
                break
        return df.reset_index(drop=True)


class Network(LinkedIn):
    def __init__(self, cookies, headers):
        LinkedIn.__init__(self)
        self.cookies = cookies
        self.headers = headers

    def get_followers(self, start=0, count=100, limit=1000):
        """
        Return an dataframe object with 13 columns:
        - PROFILE_ID                    object
        - PROFILE_URL                   object
        - PUBLIC_ID                     object
        - FIRSTNAME                     object
        - LASTNAME                      object
        - FULLNAME                      object
        - OCCUPATION                    object
        - PROFILE_PICTURE               object
        - BACKGROUND_PICTURE            object
        - FOLLOWER_COUNT                int64
        - FOLLOWING                     bool
        - INFLUENCER                    bool
        - DATE_EXTRACT                  object

        Parameters
        ----------
        start: int (default 0):
            Starting requests to LinkedIn API.

        count: int (default 100, max 100):
            Number of requests sent to LinkedIn API.

        limit: int (default 1000, unlimited=-1):
            Number of result return by function.

        """
        limit_init = limit
        df_followers = pd.DataFrame()
        if limit != -1 and limit < count:
            count = limit
        while True:
            req_url = f"{LINKEDIN_API}/network/getFollowers?start={start}&count={count}&limit={limit}"
            res = requests.post(req_url, json=self.cookies, headers=HEADERS)
            res.raise_for_status()

            # Manage LinkedIn API errors
            LinkedIn.manage_api_error(res)

            # Get json result
            res_json = res.json()
            if len(res_json) == 0:
                break
            df = pd.DataFrame(res_json)
            df_followers = pd.concat([df_followers, df], axis=0)
            start += count
            if limit != -1 and start >= limit:
                break
            time.sleep(TIME_SLEEP)
        if len(df_followers) > 0:
            df_followers = df_followers.drop_duplicates("PROFILE_ID").reset_index(
                drop=True
            )
            if limit != -1:
                df_followers = df_followers[:limit_init]
        return df_followers.reset_index(drop=True)

    def get_connections(self, start=0, count=100, limit=1000):
        """
        Return an dataframe object with 9 columns:
        - FIRSTNAME                     object
        - LASTNAME                      object
        - OCCUPATION                    object
        - CREATED_AT                    object
        - PROFILE_URL                   object
        - PROFILE_PICTURE               object
        - PROFILE_ID                    object
        - PUBLIC_ID                     object
        - DATE_EXTRACT                  object

        Parameters
        ----------
        start: int (default 0):
            Starting requests to LinkedIn API.

        count: int (default 100, max 100):
            Number of requests sent to LinkedIn API.

        limit: int (default 1000, unlimited=-1):
            Number of result return by function.

        """
        df_connections = pd.DataFrame()
        if limit != -1 and limit < count:
            count = limit
        while True:
            req_url = f"{LINKEDIN_API}/network/getConnections?start={start}&count={count}&limit={limit}"
            res = requests.post(req_url, json=self.cookies, headers=HEADERS)
            res.raise_for_status()

            # Manage LinkedIn API errors
            LinkedIn.manage_api_error(res)

            # Get json result
            res_json = res.json()
            if len(res_json) == 0:
                break
            df = pd.DataFrame(res_json)
            df_connections = pd.concat([df_connections, df], axis=0)
            start += count
            if limit != -1 and start >= limit:
                break
            time.sleep(TIME_SLEEP)
        df_connections = (
            df_connections.drop_duplicates()
            .sort_values(by="CREATED_AT", ascending=False)
            .astype(str)
        )
        return df_connections.reset_index(drop=True)


class Invitation(LinkedIn):
    def __init__(self, cookies, headers):
        LinkedIn.__init__(self)
        self.cookies = cookies
        self.headers = headers

    def get_received(self, start=0, count=100, limit=-1):
        """
        Return an dataframe object with 16 columns:
        - PROFILE_ID
        - PROFILE_URL
        - PUBLIC_ID
        - FIRSTNAME
        - LASTNAME
        - FULLNAME
        - OCCUPATION
        - PROFILE_PICTURE
        - MESSAGE
        - UNSEEN
        - SENT_AT
        - INVITATION_TYPE
        - INVITATION_DESC
        - INVITATION_STATUS
        - INVITATION_ID
        - SHARED_SECRET

        Parameters
        ----------
        start: int (default 0):
            Starting requests to LinkedIn API.

        count: int (default 100, max 100):
            Number of requests sent to LinkedIn API.

        limit: int (default -1, unlimited=-1):
            Number of result return by function.

        """
        df = pd.DataFrame()
        while True:
            if limit != -1 and limit < count:
                count = limit
            req_url = f"{LINKEDIN_API}/invitation/get?start={start}&count={count}"
            res = requests.post(req_url, json=self.cookies, headers=HEADERS)
            res.raise_for_status()

            # Manage LinkedIn API errors
            LinkedIn.manage_api_error(res)

            # Get json result
            res_json = res.json()
            if len(res_json) == 0:
                break
            tmp_df = pd.DataFrame(res_json)
            df = pd.concat([df, tmp_df], axis=0)
            start += count
            if limit != -1:
                limit -= count
            time.sleep(TIME_SLEEP)
        return df.reset_index(drop=True)

    def get_sent(self, start=0, count=100, limit=-1):
        """
        Return an dataframe object with 15 columns:
        - PROFILE_ID                    object
        - PROFILE_URL                   object
        - PUBLIC_ID                     object
        - FIRSTNAME                     object
        - LASTNAME                      object
        - FULLNAME                      object
        - OCCUPATION                    object
        - PROFILE_PICTURE               object
        - MESSAGE                       object
        - SENT_AT                       object
        - INVITATION_TYPE               object
        - INVITATION_DESC               object
        - INVITATION_STATUS             object
        - INVITATION_ID                 object
        - DATE_EXTRACT                  object

        Parameters
        ----------
        start: int (default 0):
            Starting requests to LinkedIn API.

        count: int (default 100, max 100):
            Number of requests sent to LinkedIn API.

        limit: int (default -1, unlimited=-1):
            Number of result return by function.

        """
        df = pd.DataFrame()
        while True:
            if limit != -1 and limit < count:
                count = limit
            req_url = f"{LINKEDIN_API}/invitation/getSent?start={start}&count={count}"
            res = requests.post(req_url, json=self.cookies, headers=HEADERS)
            res.raise_for_status()

            # Manage LinkedIn API errors
            LinkedIn.manage_api_error(res)

            # Get json result
            res_json = res.json()
            if len(res_json) == 0:
                break
            tmp_df = pd.DataFrame(res_json)
            df = pd.concat([df, tmp_df], axis=0)
            start += count
            if limit != -1:
                limit -= count
            time.sleep(TIME_SLEEP)
        return df.reset_index(drop=True)

    def response(
        self,
        action="accept",
        invitation_id=None,
        invitation_shared_secret=None,
        is_generic=False,
    ):
        """
        Print result : "Accept" or "Ignore"
        Return dataframe profile if invitation type = "Profile"

        Parameters
        ----------
        action: str (default 'accept')
            "accept" or "ignore"

        invitation_id: str (default None)
            Argument given in invitations.get : "INVITATION_ID"

        invitation_shared_secret: str (default None)
            Argument given in invitations.get : "SHARED_SECRET"

        is_generic: boolean (default False):
            Must be True for generic invitation, if "INVITATION_TYPE" != "Profile"
        """
        params = {
            "action": action,
            "invitation_id": invitation_id,
            "invitation_shared_secret": invitation_shared_secret,
            "is_generic": is_generic,
        }
        req_url = f"{LINKEDIN_API}/invitation/response?{urllib.parse.urlencode(params, safe='(),')}"
        res = requests.post(req_url, json=self.cookies, headers=HEADERS)
        res.raise_for_status()
        res_json = res.json()
        if action == "accept":
            print("🤝 Invitation accepted !")
        elif action == "ignore":
            print("❌ Invitation ignored !")
        return pd.DataFrame(res_json)

    def accept(
        self, invitation_id=None, invitation_shared_secret=None, is_generic=False
    ):
        """
        Return dataframe profile if invitation type = "Profile"

        Parameters
        ----------
        invitation_id: str (default None)
            Argument given in invitations.get : "INVITATION_ID"

        invitation_shared_secret: str (default None)
            Argument given in invitations.get : "SHARED_SECRET"

        is_generic: boolean (default False):
            Must be True for generic invitation, if "INVITATION_TYPE" != "Profile"
        """
        return self.response(
            "accept", invitation_id, invitation_shared_secret, is_generic
        )

    def ignore(
        self, invitation_id=None, invitation_shared_secret=None, is_generic=False
    ):
        """
        Return dataframe profile if invitation type = "Profile"

        Parameters
        ----------
        invitation_id: str (default None)
            Argument given in invitations.get : "INVITATION_ID"

        invitation_shared_secret: str (default None)
            Argument given in invitations.get : "SHARED_SECRET"

        is_generic: boolean (default False):
            Must be True for generic invitation, if "INVITATION_TYPE" != "Profile"
        """
        return self.response(
            "ignore", invitation_id, invitation_shared_secret, is_generic
        )

    def send(self, recipient_url=None, message=None):
        """
        Return res.json()

        Parameters
        ----------
        recipient_url: str: (default None)
            Profile url from Linkedin.
            Example : "https://www.linkedin.com/in/florent-ravenel/"

        message: str: (default None)
            Message send with invitation. Must not exceed characters 300 characters.
            Example : "Hi, \nI will be happy to connect with you."
        """
        if recipient_url is not None:
            recipient_id = self.get_profile_urn(recipient_url)
        # Set payload
        payload = {"inviteeProfileUrn": f"urn:li:fsd_profile:{recipient_id}"}
        # Add message in your invitations
        if message:
            # Validating message length (max size is 300 characters)
            if len(message) < 300:
                payload["customMessage"] = message
            else:
                raise BaseException(
                    f"Message too long ({len(message)} characters). Max size is 300 characters"
                )
        # Post request
        req_url = "https://www.linkedin.com/voyager/api/voyagerRelationshipsDashMemberRelationships?action=verifyQuotaAndCreate"
        res = requests.post(
            req_url,
            data=json.dumps(payload),
            cookies=self.cookies,
            headers=self.headers,
        )
        res.raise_for_status()
        time.sleep(TIME_SLEEP)
        return "✉️ Invitation successfully sent !"


class Message(LinkedIn):
    def __init__(self, cookies, headers):
        LinkedIn.__init__(self)
        self.cookies = cookies
        self.headers = headers

    def get_conversations(self, limit=-1, count=20):
        limit_max = 500
        params = {
            "limit": limit_max if limit > limit_max or limit == -1 else limit,
            "count": count,
        }
        df_result = None
        while True:
            if limit != -1:
                limit -= limit_max
                if limit < 0:
                    limit = 0
            req_url = f"{LINKEDIN_API}/message/getConversations?{urllib.parse.urlencode(params, safe='(),')}"
            res = requests.post(req_url, json=self.cookies, headers=HEADERS)
            res.raise_for_status()

            # Manage LinkedIn API errors
            LinkedIn.manage_api_error(res)

            # Get json result
            res_json = res.json()
            df = pd.DataFrame(res_json)
            created_before = (
                (int)(
                    datetime.strptime(
                        df["LAST_ACTIVITY"].iloc[-1], DATETIME_FORMAT
                    ).timestamp()
                )
            ) * 1000
            params["created_before"] = created_before
            if df_result is None:
                df_result = df
            else:
                df_result = pd.concat([df_result, df])
            time.sleep(TIME_SLEEP)
            if limit == 0 or len(df) < params["limit"]:
                break
        return df_result.reset_index(drop=True)

    def get_messages(
        self, conversation_url=None, conversation_urn=None, start=0, limit=-1, count=20
    ):
        req_url = f"{LINKEDIN_API}/message/getMessages?start={start}&limit={limit}&count{count}"
        if conversation_url:
            req_url += f"&conversation_url={conversation_url}"
        if conversation_urn:
            req_url += f"&conversation_urn={conversation_urn}"
        res = requests.post(req_url, json=self.cookies, headers=HEADERS)
        res.raise_for_status()

        # Manage LinkedIn API errors
        LinkedIn.manage_api_error(res)

        # Get json result
        res_json = res.json()
        df = pd.DataFrame(res_json)
        time.sleep(TIME_SLEEP)
        return df.reset_index(drop=True)

    def send(self, content, recipients_url=None, recipients_urn=None):
        recipient_errors = []
        params = {"action": "create"}
        message_event = {
            "eventCreate": {
                "value": {
                    "com.linkedin.voyager.messaging.create.MessageCreate": {
                        "body": content,
                        "attachments": [],
                        "attributedBody": {
                            "text": content,
                            "attributes": [],
                        },
                        "mediaAttachments": [],
                    }
                }
            }
        }
        if type(recipients_url) is str:
            recipients_url = [recipients_url]
        if type(recipients_urn) is str:
            recipients_urn = [recipients_urn]
        if type(recipients_url) is list:
            recipients_urn = []
            for recipient in recipients_url:
                recipient_urn = LinkedIn.get_profile_urn(self, recipient)
                if type(recipient_urn) is requests.exceptions.HTTPError:
                    recipient_errors.append(recipient_urn)
                recipients_urn.append(recipient_urn)
        if len(recipient_errors) > 0:
            return recipient_errors
        message_event["recipients"] = recipients_urn
        message_event["subtype"] = "MEMBER_TO_MEMBER"
        payload = {
            "keyVersion": "LEGACY_INBOX",
            "conversationCreate": message_event,
        }
        res = requests.post(
            "https://www.linkedin.com/voyager/api/messaging/conversations",
            params=params,
            json=payload,
            cookies=self.cookies,
            headers=self.headers,
        )
        time.sleep(TIME_SLEEP)
        try:
            res.raise_for_status()
            return "💬 Message successfully sent !"
        except requests.HTTPError as e:
            return e


class Post(LinkedIn):
    def __init__(self, cookies, headers):
        LinkedIn.__init__(self)
        self.cookies = cookies
        self.headers = headers

    def get_stats(self, post_url=None, activity_id=None):
        """
        Return an dataframe object with 30 columns:
        - ACTIVITY_ID                   object
        - AUTHOR_NAME                   object
        - AUTHOR_URL                    object
        - SUBDESCRIPTION                object
        - TITLE                         object
        - TEXT                          object
        - CHARACTER_COUNT               int64
        - TAGS                          object
        - TAGS_COUNT                    int64
        - EMOJIS                        object
        - EMOJIS_COUNT                  int64
        - LINKS                         object
        - LINKS_COUNT                   int64
        - PROFILE_MENTION               object
        - COMPANY_MENTION               object
        - CONTENT                       object
        - CONTENT_TITLE                 object
        - CONTENT_URL                   object
        - CONTENT_ID                    object
        - IMAGE_URL                     object
        - POLL_ID                       object
        - POLL_QUESTION                 object
        - POLL_RESULTS                  object
        - POST_URL                      object
        - VIEWS                         int64
        - COMMENTS                      int64
        - LIKES                         int64
        - SHARES                        int64
        - ENGAGEMENT_SCORE              int64
        - DATE_EXTRACT                  object

        Parameters
        ----------
        post_url: str:
            Post url from Linkedin.
            Example : "https://www.linkedin.com/posts/j%C3%A9r%C3%A9my-ravenel-8a396910_"
                      "thoughts-monday-work-activity-6891437034473426945-OOOg"

        activity_id: str (default None):
            Linkedin unique post id identifier
            Example : "6891437034473426945"

        """
        # Get profile
        if activity_id is None:
            activity_id = LinkedIn.get_activity_id(post_url)
            if activity_id is None:
                return "Please enter a valid post_url or activity_id"
        req_url = f"{LINKEDIN_API}/post/getStats?activity_id={activity_id}"
        res = requests.post(req_url, json=self.cookies, headers=HEADERS)
        res.raise_for_status()

        # Manage LinkedIn API errors
        LinkedIn.manage_api_error(res)

        # Get json result
        return pd.DataFrame(res.json()).reset_index(drop=True)

    def get_polls(self, post_url=None, activity_id=None):
        """
        Return an dataframe object with 14 columns:
        - PROFILE_ID                    object
        - PROFILE_URL                   object
        - PUBLIC_ID                     object
        - FIRSTNAME                     object
        - LASTNAME                      object
        - FULLNAME                      object
        - OCCUPATION                    object
        - PROFILE_PICTURE               object
        - BACKGROUND_PICTURE            object
        - POLL_ID                       object
        - POLL_QUESTION                 object
        - POLL_RESULT                   object
        - POST_URL                      object
        - DATE_EXTRACT                  object

        Parameters
        ----------
        post_url: str:
            Post url from Linkedin.
            Example : "https://www.linkedin.com/posts/j%C3%A9r%C3%A9my-ravenel-8a396910_"
                      "thoughts-monday-work-activity-6891437034473426945-OOOg"

        activity_id: str (default None):
            Linkedin unique post id identifier
            Example : "6891437034473426945"

        """
        # Get profile
        if activity_id is None:
            activity_id = LinkedIn.get_activity_id(post_url)
            if activity_id is None:
                return "Please enter a valid post_url or activity_id"
        req_url = f"{LINKEDIN_API}/post/getPolls?activity_id={activity_id}"
        res = requests.post(req_url, json=self.cookies, headers=HEADERS)
        res.raise_for_status()

        # Manage LinkedIn API errors
        LinkedIn.manage_api_error(res)

        # Get json result
        return pd.DataFrame(res.json()).reset_index(drop=True)

    def get_comments(
        self, post_url=None, activity_id=None, start=0, count=100, limit=-1, sleep=True
    ):
        """
        Return an dataframe object with 18 columns:
        - PROFILE_ID                    object
        - PROFILE_URL                   object
        - PUBLIC_ID                     object
        - FIRSTNAME                     object
        - LASTNAME                      object
        - FULLNAME                      object
        - OCCUPATION                    object
        - PROFILE_PICTURE               object
        - BACKGROUND_PICTURE            object
        - PROFILE_TYPE                  object
        - TEXT                          object
        - CREATED_TIME                  object
        - LANGUAGE                      object
        - DISTANCE                      object
        - COMMENTS                      int64
        - LIKES                         int64
        - POST_URL                      object
        - DATE_EXTRACT                  object

        Parameters
        ----------
        post_url: str:
            Post url from Linkedin.
            Example : "https://www.linkedin.com/posts/j%C3%A9r%C3%A9my-ravenel-8a396910_"
                      "thoughts-monday-work-activity-6891437034473426945-OOOg"

        activity_id: str (default None):
            Linkedin unique post id identifier
            Example : "6891437034473426945"

        start: int (default 0):
            Number of requests sent to LinkedIn API.
            (!) If count > 1, published date will not be returned.

        count: int (default 1, max 100):
            Number of requests sent to LinkedIn API.
            (!) If count > 1, followed at will not be returned.

        limit: int (default 10, unlimited=-1):
            Number of followers return by function. It will start with the most recent followers.

        sleep: boolean (default True):
            Sleeping time between function will be randomly between 3 to 5 seconds.

        """
        # Get profile
        if activity_id is None:
            activity_id = LinkedIn.get_activity_id(post_url)
            if activity_id is None:
                return "Please enter a valid post_url or activity_id"
        df = pd.DataFrame()
        while True:
            if limit != -1 and limit < count:
                count = limit
            req_url = f"{LINKEDIN_API}/post/getComments?activity_id={activity_id}&start={start}&count={count}"
            res = requests.post(req_url, json=self.cookies, headers=HEADERS)
            res.raise_for_status()

            # Manage LinkedIn API errors
            LinkedIn.manage_api_error(res)

            # Get json result
            res_json = res.json()
            if len(res_json) == 0:
                break
            tmp_df = pd.DataFrame(res_json)
            df = pd.concat([df, tmp_df], axis=0)
            start += count
            if limit != -1:
                limit -= count
            if sleep:
                time.sleep(TIME_SLEEP)
        return df.reset_index(drop=True)

    def get_likes(
        self, post_url=None, activity_id=None, start=0, count=100, limit=-1, sleep=True
    ):
        """
        Return an dataframe object with 13 columns:
        - PROFILE_ID                    object
        - PROFILE_URL                   object
        - PUBLIC_ID                     object
        - FIRSTNAME                     object
        - LASTNAME                      object
        - FULLNAME                      object
        - OCCUPATION                    object
        - PROFILE_PICTURE               object
        - BACKGROUND_PICTURE            object
        - PROFILE_TYPE                  object
        - REACTION_TYPE                 object
        - POST_URL                      object
        - DATE_EXTRACT                  object

        Parameters
        ----------
        post_url: str:
            Post url from Linkedin.
            Example : "https://www.linkedin.com/posts/j%C3%A9r%C3%A9my-ravenel-8a396910_"
                      "thoughts-monday-work-activity-6891437034473426945-OOOg"

        activity_id: str (default None):
            Linkedin unique post id identifier
            Example : "6891437034473426945"

        start: int (default 0):
            Number of requests sent to LinkedIn API.
            (!) If count > 1, published date will not be returned.

        count: int (default 1, max 100):
            Number of requests sent to LinkedIn API.
            (!) If count > 1, followed at will not be returned.

        limit: int (default 10, unlimited=-1):
            Number of followers return by function. It will start with the most recent followers.

        sleep: boolean (default True):
            Sleeping time between function will be randomly between 3 to 5 seconds.

        """
        # Get profile
        if activity_id is None:
            activity_id = LinkedIn.get_activity_id(post_url)
            if activity_id is None:
                return "Please enter a valid post_url or activity_id"
        df = pd.DataFrame()
        while True:
            if limit != -1 and limit < count:
                count = limit
            req_url = f"{LINKEDIN_API}/post/getLikes?activity_id={activity_id}&start={start}&count={count}"
            res = requests.post(req_url, json=self.cookies, headers=HEADERS)
            res.raise_for_status()

            # Manage LinkedIn API errors
            LinkedIn.manage_api_error(res)

            # Get json result
            res_json = res.json()
            if len(res_json) == 0:
                break
            tmp_df = pd.DataFrame(res_json)
            df = pd.concat([df, tmp_df], axis=0)
            start += count
            if limit != -1:
                limit -= count
            if sleep:
                time.sleep(TIME_SLEEP)
        return df.reset_index(drop=True)


class Event(LinkedIn):
    def __init__(self, cookies, headers):
        LinkedIn.__init__(self)
        self.cookies = cookies
        self.headers = headers

    def get_guests(
        self, event_url="https://www.linkedin.com/events/6762355783188525056/"
    ):
        """
        Return an dataframe object with 8 columns:
        - FULLNAME                      object
        - PROFILE_ID                    object
        - PROFILE_URL                   object
        - PUBLIC_ID                     object
        - OCCUPATION                    object
        - LOCATION                      object
        - DISTANCE                      object
        - DATE_EXTRACT                  object

        Parameters
        ----------
        event_url: str:
            Event url from Linkedin.
            Example : "https://www.linkedin.com/events/6762355783188525056/"

        """
        req_url = f"{LINKEDIN_API}/event/getGuests?event_link={event_url}"
        res = requests.post(req_url, json=self.cookies, headers=HEADERS)
        res.raise_for_status()
        return pd.DataFrame(res.json()).reset_index(drop=True)


class Company(LinkedIn):
    def __init__(self, cookies, headers):
        LinkedIn.__init__(self)
        self.cookies = cookies
        self.headers = headers

    def get_info(self, company_url="https://www.linkedin.com/company/naas-ai/"):
        """
        Return an dataframe object with 18 columns:
        - COMPANY_ID                    object
        - COMPANY_URL                   object
        - COMPANY_NAME                  object
        - UNIVERSAL_NAME                object
        - LOGO_URL                      object
        - INDUSTRY_ID                   object
        - INDUSTRY                      object
        - WEBSITE                       object
        - TAGLINE                       object
        - SPECIALITIES                  object
        - DESCRIPTION                   object
        - COUNTRY                       object
        - REGION                        object
        - CITY                          object
        - STAFF_COUNT                   object
        - STAFF_RANGE                   object
        - FOLLOWER_COUNT                object
        - DATE_EXTRACT                  object

        Parameters
        ----------
        company_url: str:
            Company url from Linkedin.
            Example : "https://www.linkedin.com/company/naas-ai/"

        """
        df = pd.DataFrame()
        req_url = f"{LINKEDIN_API}/company/getInfo?company_url={company_url}"
        res = requests.post(req_url, json=self.cookies, headers=HEADERS)
        res.raise_for_status()

        # Manage LinkedIn API errors
        LinkedIn.manage_api_error(res)

        # Get json result
        df = pd.DataFrame(res.json())
        return df.reset_index(drop=True)

    def get_followers(
        self,
        company_url="https://www.linkedin.com/company/naas-ai/",
        start=0,
        count=1,
        limit=10,
        sleep=True,
    ):
        """
        Return an dataframe object with 11 columns:
        - FIRSTNAME                     object
        - LASTNAME                      object
        - OCCUPATION                    object
        - PROFILE_PICTURE               object
        - PROFILE_URL                   object
        - PROFILE_ID                    object
        - PUBLIC_ID                     object
        - FOLLOWED_AT                   object
        - DISTANCE                      object
        - DATE_EXTRACT                  object

        Parameters
        ----------
        company_url: str:
            Company url from Linkedin.
            Example : "https://www.linkedin.com/company/naas-ai/"

        start: int (default 0):
            Number of requests sent to LinkedIn API.
            (!) If count > 1, published date will not be returned.

        count: int (default 1, max 100):
            Number of requests sent to LinkedIn API.
            (!) If count > 1, followed at will not be returned.

        limit: int (default 10, unlimited=-1):
            Number of followers return by function. It will start with the most recent followers.

        sleep: boolean (default True):
            Sleeping time between function will be randomly between 3 to 5 seconds.

        """
        df = pd.DataFrame()
        while True:
            if limit != -1 and limit < count:
                count = limit
            req_url = f"{LINKEDIN_API}/company/getFollowers?company_url={company_url}&start={start}&count={count}"
            res = requests.post(req_url, json=self.cookies, headers=HEADERS)
            res.raise_for_status()

            # Manage LinkedIn API errors
            LinkedIn.manage_api_error(res)

            # Get json result
            res_json = res.json()
            if len(res_json) == 0:
                break
            tmp_df = pd.DataFrame(res_json)
            df = pd.concat([df, tmp_df], axis=0)
            start += count
            if limit != -1:
                limit -= count
            if sleep:
                time.sleep(TIME_SLEEP)
        if len(df) > 0:
            df = df.sort_values(by="FOLLOWED_AT", ascending=False)
        return df.reset_index(drop=True)

    def __get_posts_views(self, activity_id):
        views = 0
        req_url = f"{LINKEDIN_API}/company/getPostsViews?activity_id={activity_id}"
        res = requests.post(req_url, json=self.cookies, headers=HEADERS)
        res.raise_for_status()

        # Manage LinkedIn API errors
        LinkedIn.manage_api_error(res)

        # Get result
        views = res.json().get("VIEWS")
        time.sleep(3)
        return views

    def get_posts_feed(
        self,
        company_url,
        start=0,
        count=100,
        limit=-1,
        sleep=True,
    ):
        """
        Return an dataframe object with 31 columns:
        - ACTIVITY_ID                   object
        - PUBLISHED_DATE                object
        - AUTHOR_NAME                   object
        - AUTHOR_URL                    object
        - SUBDESCRIPTION                object
        - TITLE                         object
        - TEXT                          object
        - CHARACTER_COUNT               int64
        - TAGS                          object
        - TAGS_COUNT                    int64
        - EMOJIS                        object
        - EMOJIS_COUNT                  int64
        - LINKS                         object
        - LINKS_COUNT                   int64
        - PROFILE_MENTION               object
        - COMPANY_MENTION               object
        - CONTENT                       object
        - CONTENT_TITLE                 object
        - CONTENT_URL                   object
        - CONTENT_ID                    object
        - IMAGE_URL                     object
        - POLL_ID                       object
        - POLL_QUESTION                 object
        - POLL_RESULTS                  object
        - POST_URL                      object
        - VIEWS                         int64
        - COMMENTS                      int64
        - LIKES                         int64
        - SHARES                        int64
        - ENGAGEMENT_SCORE              float64
        - DATE_EXTRACT                  object

        Parameters
        ----------
        company_url: str:
            Company url from Linkedin.
            Example : "https://www.linkedin.com/company/naas-ai/"

        start: int (default 0):
            Starting request sent to LinkedIn API.

        count: int (default 1, max 100):
            Number of requests sent to LinkedIn API.
            (!) If count > 1, published date will not be returned.

        limit: int (default 10, unlimited=-1):
            Number of posts return by function. It will start with the most recent post.

        sleep: boolean (default True):
            Sleeping time between function will be randomly between 5 to 10 seconds.

        """
        # Loop init
        df = pd.DataFrame()
        if limit != -1 and count > limit:
            count = limit
        while True:
            req_url = f"{LINKEDIN_API}/company/getPostsFeed?company_url={company_url}&start={start}&count={count}"
            res = requests.post(req_url, json=self.cookies, headers=HEADERS)
            res.raise_for_status()

            # Manage LinkedIn API errors
            LinkedIn.manage_api_error(res)

            # Get json result
            res_json = res.json()
            if len(res_json) == 0:
                break
            # Get response in dataframe
            tmp_df = pd.DataFrame(res_json)

            # Concat dataframe
            df = pd.concat([df, tmp_df], axis=0)

            # Time sleep to avoid linkedin ban
            if sleep:
                time.sleep(TIME_SLEEP)
            start += count
            if limit != -1 and start >= limit:
                break
        # Cleaning
        if len(df) > 0:
            # Add views + engagement score
            df["VIEWS"] = df.apply(
                lambda row: self.__get_posts_views(row.ACTIVITY_ID), axis=1
            )
            df["ENGAGEMENT_SCORE"] = 0
            df.loc[df["VIEWS"] != 0, "ENGAGEMENT_SCORE"] = (
                df["COMMENTS"] + df["LIKES"]
            ) / df["VIEWS"]
        return df.reset_index(drop=True)



import pandas as pd
import datetime as dt


class Yahoofinance:
    def get(
        self,
        tickers,
        date_from=-36,
        date_to="today",
        interval="1d",
        moving_averages: list = [],
        moving_average_col="Close",
    ):
        """Generate financial data"""
        # Init dataframe
        df_stocks = pd.DataFrame()

        # If tickers is string => change to list
        if isinstance(tickers, str):
            tickers = [tickers]
        # Set date from
        if isinstance(date_from, int):
            if date_from < 0:
                date_from = dt.datetime.today() + dt.timedelta(days=date_from)
            else:
                error_text = f"❌ date_from ({date_from}) cannot be positive."
                print(error_text)
        # Set date to
        if isinstance(date_to, int):
            if date_to < 0:
                date_to = dt.datetime.today() + dt.timedelta(days=date_to)
            else:
                error_text = f"❌ date_from ({date_to}) cannot be positive."
                print(error_text)
        if isinstance(date_to, str):
            if date_to == "today":
                date_to = dt.datetime.today()
            else:
                error_text = f"❌ date_to cannot be ({date_to}). Try with 'today' or use an integer."
                print(error_text)
        period1 = date_from.strftime("%s")
        period2 = date_to.strftime("%s")
        for ticker in tickers:
            url = (
                f"https://query1.finance.yahoo.com/v7/finance/download/"
                f"{ticker}?period1={period1}&period2={period2}&interval={interval}&events=history"
            )
            df = pd.read_csv(url)
            df["Ticker"] = ticker
            if len(moving_averages) > 0:
                if moving_average_col in df.columns:
                    for moving_average in moving_averages:
                        if isinstance(moving_average, str):
                            moving_average = int(moving_average)
                        if isinstance(moving_average, int):
                            df[f"MA{moving_average}"] = (
                                df[moving_average_col].rolling(moving_average).mean()
                            )
                        else:
                            error_text = f"❌ Moving average '{moving_average_col}' not recognize."
                            print(error_text)
                else:
                    error_text = f"❌ We can not calculate moving averages. Columns '{moving_average_col}' does not exist in dataframe."
                    print(error_text)
            df_stocks = pd.concat([df_stocks, df])
        df_stocks["Date"] = pd.to_datetime(df_stocks["Date"], format="%Y-%m-%d")
        return df_stocks



from naas_drivers.driver import InDriver
import pandas as pd
import requests
import uuid
import os
import warnings
from IPython.core import display
from htmlBuilder import tags, attributes

#  https://litmus.com/community/templates/31-accessible-product-announcement-email
# https://github.com/rodriguezcommaj/accessible-emails
base_style = """
/* CLIENT-SPECIFIC STYLES */
body, table, td, a { -webkit-text-size-adjust: 100%; -ms-text-size-adjust: 100%; }
table, td { mso-table-lspace: 0pt; mso-table-rspace: 0pt; }
img { -ms-interpolation-mode: bicubic; }

/* RESET STYLES */
img { border: 0; height: auto; line-height: 100%; outline: none; text-decoration: none; }
table { border-collapse: collapse !important; text-align: left !important; }
body { height: 100% !important; margin: 0 !important; padding: 0 !important; width: 100% !important; }

/* iOS BLUE LINKS */
a[x-apple-data-detectors] {
    color: inherit !important;
    text-decoration: none !important;
    font-size: inherit !important;
    font-family: inherit !important;
    font-weight: inherit !important;
    line-height: inherit !important;
}

/* GMAIL BLUE LINKS */
u + #body a {
    color: inherit;
    text-decoration: none;
    font-size: inherit;
    font-family: inherit;
    font-weight: inherit;
    line-height: inherit;
}

/* SAMSUNG MAIL BLUE LINKS */
#MessageViewBody a {
    color: inherit;
    text-decoration: none;
    font-size: inherit;
    font-family: inherit;
    font-weight: inherit;
    line-height: inherit;
}

a.link { color: #B200FD; font-weight: 600; text-decoration: underline; }
a.link:hover { color: #000000 !important; text-decoration: none !important; background-color: #5c1958 !important; }
a.button:hover { color: #ffffff !important; background-color: #5c1958 !important; }

td, th {
    padding: 10px;
}

th {
    text-align: center;
}

table * {
    margin: 18px 0 !important;
    font-size: 18px;
}

.table_border {
  border-collapse: collapse;
  overflow: hidden;
}
.table_border tr:hover {
  background-color: AliceBlue !important;
  color: black;
}

.table_border tr:nth-child(even) { background-color: ghostwhite}

@media screen and (min-width:600px) {
    h1 { font-size: 48px !important; line-height: 48px !important; }
    .intro { font-size: 24px !important; line-height: 36px !important; }
}
.basic_font {
    font-family: 'Avenir Next', -apple-system, BlinkMacSystemFont, 'Segoe UI', Roboto, Helvetica, Arial, sans-serif, 'Apple Color Emoji', 'Segoe UI Emoji', 'Segoe UI Symbol';  # noqa: E501
}
"""

table_ie9 = """
<!--[if (gte mso 9)|(IE)]>
<table cellspacing="0" cellpadding="0" border="0" width="720" align="center" role="presentation"><tr><td>
<![endif]-->
"""

table_ie9_close = """
<!--[if (gte mso 9)|(IE)]>
</td></tr></table>
<![endif]-->
"""

# Link Follow us
GIT_LINK = "https://github.com/jupyter-naas/naas"
YT_LINK = "https://www.youtube.com/channel/UCKKG5hzjXXU_rRdHHWQ8JHQ?sub_confirmation=1"
LK_LINK = "https://www.linkedin.com/showcase/naas-ai"
TW_LINK = "https://twitter.com/JupyterNaas"

# Logo Follow us
GIT_IMG_BLACK = "https://icons.iconarchive.com/icons/icons8/windows-8/512/Programming-Github-icon.png"
YT_IMG_BLACK = "https://icons.iconarchive.com/icons/icons8/windows-8/512/Social-Networks-Youtube-icon.png"
LK_IMG_BLACK = "https://icons.iconarchive.com/icons/icons8/windows-8/512/Social-Networks-Linkedin-icon.png"
TW_IMG_BLACK = "https://icons.iconarchive.com/icons/icons8/windows-8/512/Social-Networks-Twitter-icon.png"


class Align(attributes.HtmlTagAttribute):
    """Specifies alignement position"""


class EmailBuilder(InDriver):
    """EmailBuilder generator lib"""

    deprecated = False

    def __init__(self, deprecated=False):
        self.deprecated = deprecated

    def deprecatedPrint(self, function=None):
        # TODO remove this in june 2021
        if self.deprecated:
            warnings.warn(
                """[Warning], naas_drivers.html or naas_drivers.emailBuilder is deprecated,
                use naas_drivers.emailbuilder instead it will be remove in 1 june 2021"""
            )
        if function:
            # TODO remove this in July 2021
            warnings.warn(
                f"""[Warning], The function emailbuilder.{function}() is deprecated,
                it will be remove in 1 July 2021"""
            )

    def __align(self, mode):
        margin = "0 auto 0 0"
        if mode == "right":
            margin = "0 0 0 auto"
        if mode == "center":
            margin = "0 auto 0 auto"
        return margin

    def __convert(self, data, name):
        args = [data]
        if name.find("_") > -1:
            args_list = name.split("_")[1:]
            args_list = list(filter(lambda x: not x.isdigit(), args_list))
            args.extend(args_list)
        if name.startswith("image") and isinstance(data, str):
            return self.image(*args)
        if name.startswith("logo") and isinstance(data, str):
            return self.logo(*args)
        elif name.startswith("link") and isinstance(data, str):
            return self.link(*args)
        elif name.startswith("button") and isinstance(data, str):
            return self.button(*args)
        elif name.startswith("info") and isinstance(data, str):
            return self.button(*args)
        elif name.startswith("title") and isinstance(data, str):
            return self.title(*args)
        elif name.startswith("heading") and isinstance(data, str):
            return self.heading(*args)
        elif name.startswith("address") and isinstance(data, str):
            return self.address(*args)
        elif name.startswith("table") and (
            isinstance(data, pd.DataFrame) or isinstance(data, list)
        ):
            return self.table(*args)
        elif name.startswith("subheading") and isinstance(data, str):
            return self.subheading(*args)
        else:
            return self.text(*args) if isinstance(data, str) else data

    def __create_network_icon(self, img_src, href, width, padding, margin, bg_color):
        return tags.A(
            attributes.Href(href),
            attributes.Class("link"),
            attributes.Target("_blank"),
            tags.Img(
                attributes.Width("6%"),
                Align("center"),
                attributes.Src(img_src),
                attributes.InlineStyle(
                    padding=padding,
                    margin=margin,
                    border_radius="20%",
                    background_color=bg_color,
                ),
            ),
        )

    def __default_networks(self):
        return [
            {"img_src": GIT_IMG_BLACK, "href": GIT_LINK},
            {"img_src": YT_IMG_BLACK, "href": YT_LINK},
            {"img_src": LK_IMG_BLACK, "href": LK_LINK},
            {"img_src": TW_IMG_BLACK, "href": TW_LINK},
        ]

    def __default_company(self):
        return [
            "Naas: All-in-one data platform",
            "<a target='_blank' href='https://www.naas.ai/'>www.naas.ai</a>",
        ]

    def __default_legal(self):
        return ["Powered by CASHSTORY SAS © 2022"]

    def __text_style(
        self,
        style={},
        color=None,
        font_size=None,
        text_align=None,
        bold=False,
        underline=False,
        italic=False,
    ):
        if color:
            style["color"] = color
        if font_size:
            style["font_size"] = f"{font_size}"
        if text_align:
            style["text_align"] = text_align
        if bold:
            style["font_weight"] = "bold"
        if underline:
            style["text_decoration"] = "underline"
        if italic:
            style["font_style"] = "italic"
        return style

    def __create_table_cell(
        self,
        content,
        header=False,
        align=None,
        border=False,
        header_bg_color="black",
        header_ft_color="white",
        border_color="black",
        size=None,
    ):
        style = {}
        if border:
            style["border"] = f"1px solid {border_color}"
        if size:
            style["width"] = f"{size}"
        if header:
            style["color"] = header_ft_color
            style["background_color"] = header_bg_color
            return tags.Th(content, attributes.InlineStyle(**style))
        else:
            if align is not None:
                style["text_align"] = align
            return tags.Td(content, attributes.InlineStyle(**style))

    def __df_table_header(
        self,
        df,
        border=False,
        header_bg_color="black",
        header_ft_color="white",
        border_color="black",
    ):
        res = []
        for col in df.columns:
            res.append(
                self.__create_table_cell(
                    tags.Text(col), True, None, border, header_bg_color, header_ft_color
                )
            )
        return res

    def title(
        self,
        title,
        heading=None,
        color="#000000",
        font_size="32px",
        text_align="center",
        bold=True,
        underline=False,
        italic=False,
    ):
        self.deprecatedPrint()
        style = {"font_weight": "normal", "line_height": "32px", "margin": "48px 0"}
        style = self.__text_style(
            style, color, font_size, text_align, bold, underline, italic
        )
        return tags.H1(
            attributes.InlineStyle(**style),
            tags.Text(title),
            tags.Br(),
            (
                tags.Span(
                    attributes.InlineStyle(
                        font_size="24px", font_weight="600", color="darkgray"
                    ),
                    tags.Text(heading),
                )
                if heading
                else None
            ),
        )

    def heading(
        self,
        text,
        color="#000000",
        font_size=28,
        text_align="center",
        bold=True,
        underline=False,
        italic=False,
    ):
        self.deprecatedPrint()
        style = {
            "font-weight": "normal",
            "line_height": "32px",
            "margin": "48px 0 24px 0",
            "text_align": text_align,
        }
        style = self.__text_style(
            style, color, font_size, text_align, bold, underline, italic
        )
        return tags.H2(
            attributes.InlineStyle(**style),
            tags.Text(text),
        )

    def subheading(
        self,
        text,
        color="#000000",
        font_size="24px",
        text_align="left",
        bold=False,
        underline=False,
        italic=False,
    ):
        self.deprecatedPrint()
        return self.text(text, color, font_size, text_align, bold, underline, italic)

    def text(
        self,
        text,
        color="#000000",
        font_size="18px",
        text_align=None,
        bold=False,
        underline=False,
        italic=False,
        padding_left="10px",
        padding_right="10px",
    ):
        self.deprecatedPrint()
        style = {"padding_left": padding_left, "padding_right": padding_right}
        style = self.__text_style(
            style, color, font_size, text_align, bold, underline, italic
        )
        return tags.P(
            tags.Text(text),
            attributes.InlineStyle(**style),
        )

    def link(
        self,
        link,
        title="Open",
        color="#B200FD",
        font_size="18px",
        text_align="left",
        bold=False,
        underline=True,
        italic=False,
    ):
        self.deprecatedPrint()
        style = self.__text_style(
            {}, color, font_size, text_align, bold, underline, italic
        )
        return tags.A(
            attributes.Class("link"),
            attributes.Href(link),
            attributes.InlineStyle(**style),
            tags.Text(title),
        )

    def list(self, list_):
        elems = []
        for elem in list_:
            elems.append(
                tags.Li(
                    elem
                    if issubclass(type(elem), tags.HtmlTag)
                    else tags.Text(f"{elem}")
                )
            )
        return tags.Ul(elems)

    def button(
        self,
        link,
        text="Open",
        width="auto",
        color="white",
        background_color="black",
    ):
        self.deprecatedPrint()
        return tags.Center(
            tags.Div(
                attributes.InlineStyle(margin="48px 0"),
                tags.A(
                    attributes.Class("button link"),
                    attributes.Href(link),
                    attributes.InlineStyle(
                        background_color=background_color,
                        color=color,
                        border_radius="4px",
                        display="inline-block",
                        font_family="sans-serif",
                        font_size="18px",
                        font_weight="bold",
                        line_height="60px",
                        text_align="center",
                        text_decoration="none",
                        width=width,
                        max_width="300px",
                        padding_left="10px",
                        padding_right="10px",
                        _webkit_text_size_adjust="none",
                    ),
                    tags.Text(text),
                ),
            )
        )

    def table(
        self,
        data,
        border=False,
        header=False,
        col_align={},
        header_bg_color="black",
        header_ft_color="white",
        border_color="black",
        col_size={},
    ):
        self.deprecatedPrint()
        elems = []
        table_arr = None
        row_link = False
        row_link_index = -1
        index = 0
        if isinstance(data, pd.DataFrame):
            table_arr = list(data.values.tolist())
            row_link = True if "row_link" in data.columns else False
            try:
                row_link_index = list(data.columns).index("row_link")
            except ValueError:
                pass
        elif isinstance(data, list):
            table_arr = data
        else:
            error_text = f"Table should be array, not {data.__name__}"
            self.print_error(error_text)
            return None
        for row in table_arr:
            res = []
            is_header = index == 0 and header is True
            if isinstance(row, list):
                for i in range(len(row)):
                    align = col_align[i] if i in col_align else None
                    size = col_size[i] if i in col_size else None
                    cell = row[i]

                    if isinstance(data, pd.DataFrame):
                        col = list(data.columns)[i]
                        if is_header:
                            elems.append(
                                tags.Tr(
                                    self.__df_table_header(
                                        data,
                                        border,
                                        header_bg_color,
                                        header_ft_color,
                                        border_color,
                                    )
                                )
                            )
                            is_header = False
                        if row_link:
                            if col != "row_link":
                                link = row[row_link_index]
                                res.append(
                                    self.__create_table_cell(
                                        tags.A(
                                            attributes.Class("link"),
                                            attributes.Href(link),
                                            self.__convert(cell, col),
                                        ),
                                        header=is_header,
                                        align=align,
                                        border=border,
                                        header_bg_color=header_bg_color,
                                        header_ft_color=header_ft_color,
                                        border_color=border_color,
                                    )
                                )
                        else:
                            res.append(
                                self.__create_table_cell(
                                    self.__convert(cell, col),
                                    is_header,
                                    align,
                                    border,
                                    header_bg_color,
                                    header_ft_color,
                                    border_color,
                                    size,
                                )
                            )
                    else:
                        res.append(
                            self.__create_table_cell(
                                (tags.Text(cell) if isinstance(cell, str) else cell),
                                is_header,
                                align,
                                border,
                                header_bg_color,
                                header_ft_color,
                                border_color,
                                size,
                            )
                        )
            else:
                align = col_align[0] if 0 in col_align else None
                size = col_size[0] if 0 in col_size else None
                res.append(
                    self.__create_table_cell(
                        (tags.Text(row) if isinstance(row, str) else row),
                        is_header,
                        align,
                        border,
                        header_bg_color,
                        header_ft_color,
                        border_color,
                        size,
                    )
                )
            elems.append(tags.Tr(res))
            index += 1
        tab = tags.Table(
            attributes.InlineStyle(width="100%", border_collapse="collapse"),
            attributes.Class("table_border") if border else None,
            elems,
        )
        return tags.P(tab)

    def image(
        self, src, link=None, name="Cover", align="left", width="100%", height="80%"
    ):
        self.deprecatedPrint()
        if src is None:
            return None
        elems_img = [
            attributes.Src(f"{src}?naas_uid={str(uuid.uuid4())}"),
            attributes.Height(height),
            attributes.Width(width),
            {"name": "border", "value": 0},
            attributes.InlineStyle(
                border_radius="4px",
                margin=self.__align(align),
                display="block",
            ),
            {"name": "alt", "value": name},
        ]
        if link:
            return tags.A(
                attributes.Class("link"), attributes.Href(link), tags.Img(elems_img)
            )
        else:
            return tags.Img(elems_img)

    def footer(self, text, first=None, *elems):
        self.deprecatedPrint()
        one = [
            attributes.InlineStyle(
                font_size="16px",
                font_weight="400",
                line_height="24px",
                margin_top="48px",
            ),
            tags.Text(text),
            first,
        ]
        return tags.Footer(tags.P(one), *elems)

    def footer_company(
        self,
        networks=None,
        company=None,
        legal=None,
        logo_width="6%",
        logo_padding="5px",
        logo_margin="0px 15px",
        logo_bg_color="white",
        naas=False,
    ):
        self.deprecatedPrint()
        net = []
        com = []
        leg = []
        if naas:
            networks = self.__default_networks()
            company = self.__default_company()
            # legal = self.__default_legal()
        if networks:
            for elem in networks:
                if elem.keys() >= frozenset({"img_src", "href"}):
                    net.append(
                        self.__create_network_icon(
                            **elem,
                            width=logo_width,
                            padding=logo_padding,
                            margin=logo_margin,
                            bg_color=logo_bg_color,
                        )
                    )
            net = tags.P(net, attributes.InlineStyle(text_align="center"))
        if company:
            for elem in company:
                com.append(tags.B(tags.Text(elem)))
                com.append(tags.Br())
            com = tags.P(
                attributes.InlineStyle(line_height=1, text_align="center"),
                tags.Span(com, attributes.InlineStyle(font_size="12px")),
            )
        if legal:
            for elem in legal:
                leg.append(tags.Text(elem))
                leg.append(tags.Br())
            leg = tags.P(
                attributes.InlineStyle(line_height=1, text_align="center"),
                tags.Span(leg, attributes.InlineStyle(font_size="11px")),
            )
        return tags.Footer(tags.Hr(), net, com, leg)

    def main(
        self,
        **kwargs,
    ):
        self.deprecatedPrint()
        items = []
        for key, value in kwargs.items():
            items.append(self.__convert(value, key))
        return [tags.Main(items)]

    def __display(self, content, mode):
        uid = uuid.uuid4().hex
        if mode is None:
            return
        if mode == "shadow":
            shadow = f"""
            <script type="text/javascript">
            var div_{uid} = document.createElement('div');
            var shadowRoot_{uid} = div_{uid}.attachShadow({{mode: 'open'}});
            shadowRoot_{uid}.innerHTML = `
            {content}
            `;
            console.log('injected')
            </script>
            """
            display.display(display.HTML(shadow))
        elif mode == "iframe":
            iframe = f"""
            <script>
                function resizeIframe_{uid}(obj) {{
                    obj.style.height = obj.contentWindow.document.documentElement.scrollHeight + 'px';
                }}
            </script>
            <iframe id="FileFrame_{uid}" src="about:blank"
                style="border: 0; width: 100%; height: 100%" onload="resizeIframe_{uid}(this)">
            </iframe>
            <script type="text/javascript">
            var doc_{uid} = document.getElementById('FileFrame_{uid}').contentWindow.document;
            doc_{uid}.open();
            doc_{uid}.write(`{content}`);
            doc_{uid}.close();
            </script>
            """
            display.display(display.HTML(iframe))
        elif mode == "embed":
            display.display(display.HTML(content))

    def __export(self, html, filename, css=""):
        result = html.replace("</head>", f'<style id="naas_css">{css}</style></head>')
        extension = filename.split(".")[1]
        output = None
        json = {
            "html": result,
            "emulateScreenMedia": True,
            "ignoreHttpsErrors": True,
            "scrollPage": False,
            "screenshot": {"type": extension},
        }
        if filename.endswith(".png") or filename.endswith(".jpeg"):
            output = "screenshot"
            json["screenshot"] = {"type": extension}
        elif filename.endswith(".pdf"):
            output = "pdf"
            json["pdf"] = {"width": "20.5cm", "height": "36.5cm"}
        else:
            error_text = f"extension {filename.split('.')[1]} not suported for now"
            self.print_error(error_text)
        json["output"] = output
        req = requests.post(
            url=f"{os.environ.get('SCREENSHOT_API', 'http://naas-screenshot:9000')}/api/render",  # Sensitive
            json=json,
        )
        req.raise_for_status()
        open(filename, "wb").write(req.content)
        print(f"Saved as {filename}")

    def export(self, html, filenames, css=None):
        """create html export and add css to it"""
        self.deprecatedPrint()
        if isinstance(filenames, list):
            for filename in filenames:
                self.__export(html, filename, css)
        else:
            self.__export(html, filenames, css)

    def generate(self, title=None, logo=None, display="embed", footer=None, **kwargs):
        self.deprecatedPrint()
        gen_html = tags.Html(
            attributes.Lang("en"),
            tags.Head(
                tags.Meta(
                    attributes.HttpEquiv("Content-Type"),
                    attributes.Content("text/html; charset=utf-8"),
                ),
                tags.Meta(
                    attributes.HttpEquiv("Content-Type"),
                    attributes.Content("width=device-width, initial-scale=1"),
                ),
                tags.Meta(
                    attributes.Name("viewport"),
                    attributes.Content("width=device-width, initial-scale=1"),
                ),
                tags.Meta(
                    attributes.HttpEquiv("X-UA-Compatible"),
                    attributes.Content("IE=edge"),
                ),
                tags.Style(tags.Text(base_style)),
                tags.Title(tags.Text(title)) if title else None,
            ),
            tags.Body(
                attributes.InlineStyle(margin="0 !important", padding="0 !important"),
                tags.Div(
                    attributes.InlineStyle(
                        display="none", max_height="0", overflow="hidden"
                    ),
                    tags.Text(title) if title else None,
                ),
                tags.Div(
                    attributes.InlineStyle(
                        display="none", max_height="0", overflow="hidden"
                    ),
                    tags.Text("&nbsp;‌" * 240),
                ),
                tags.Text(table_ie9),
                tags.Div(
                    {"name": "role", "value": "article"},
                    {"name": "aria-label", "value": title} if title else None,
                    attributes.Lang("en"),
                    attributes.Class("basic_font"),
                    attributes.InlineStyle(
                        background_color="white",
                        color="#2b2b2b",
                        font_size="18px",
                        font_weight="400",
                        line_height="28px",
                        margin="0 auto",
                        max_width="720px",
                        padding="40px 20px 40px 20px",
                    ),
                    self.title(title) if title else None,
                    self.main(**kwargs),
                    footer,
                ),
                tags.Text(table_ie9_close),
            ),
        )
        res = gen_html.render()
        self.__display(res, display)
        return res
